from PIL import Image, ImageSequence
from .encoders import *
from .utils import *
import time

class ImageHandler:
    """A class to handle images for cryptography

    Attributes
    ----------
    image : Image - The parsed image

    Methods
    -------
    write(file_path): Writes the image data to a file
    file_info(): Gets the file information of the image
    encode(method, **kwargs): Encodes data into the image
    decode(method, **kwargs): Decodes data from the image
    """

    def __init__(self, file_path):
        """Initialise the ImageHandler class

        :param file_path: string - The path to the image file

        :attribute image: Image - The parsed image

        :return: object - The ImageHandler object
        """

        self.image = Image.open(file_path)

    def write(self, file_path):
        """Write to the image data to a file

        :param file_path: string - The path to the file to write to
        :return: None
        """

        self.image.save(file_path)

    def file_info(self):
        """Get the file information of the image

        :return: dict - The file information
        """

        return {"mode": self.image.mode, "size": self.image.size, "format": self.image.format, "filename": self.image.filename}

    def encode(self, method, **kwargs):
        """Encode data into the image

        :param method: string - The method to use to encode the data
        :param kwargs: dict - The arguments to pass to the method

        :return: None

        :raises: NotImplementedError - If the method is not implemented

        The method allows selecting the method to use to encode the data, and
        it accepts keyword arguments to pass to the method. The keywords are:
        - data: The data to encode
        - key: The key to use to encode the data
        """
        match method:
            case "rail_fence_cipher":
                info = self.file_info()
                x = get_rail_fence_pixels(info["size"][0], info["size"][1], kwargs["key"])
                encoded_data = encode_rail_fence_cipher(kwargs["data"], kwargs["key"])
                for (idx, pixel) in enumerate(x):
                    new_pixel = encode_data_to_pixel(self.image.getpixel(pixel), encoded_data[idx])
                    self.image.putpixel(i, new_pixel)
                
            case "random_spacing":
                info = self.file_info()
                enumerator = get_random_spacing_pixels(info["size"][0], info["size"][1], kwargs["key"])
                data = kwargs["data"]
                for (idx, pixel) in enumerate(enumerator):
                    new_pixel = encode_data_to_pixel(self.image.getpixel(pixel), data[idx])
                    self.image.putpixel(pixel, new_pixel)
                


            case _:
                raise NotImplementedError(f"Method {method} not implemented")
                    

    def decode(self, method, **kwargs):
        """Decode data from the image

        :param method: string - The method to use to decode the data
        :param kwargs: dict - The arguments to pass to the method

        :return: string - The decoded data

        :raises: NotImplementedError - If the method is not implemented

        The method allows selecting the method to use to decode the data, and
        it accepts keyword arguments to pass to the method. The keywords are:
        - key: The key to use to decode the data
        """

        match method:
            case _:
                raise NotImplementedError(f"Method {method} not implemented")


image_handler = ImageHandler("img/d.png")
#image_handler.encode("lsb")
#image_handler.write("img/d.png")


import unittest
import pytest
from src.utils import encode_data_to_pixel, decode_data_from_pixel

class PixelDataEncodeTests(unittest.TestCase):
    def test_can_encode_pixel(self):

        result = encode_data_to_pixel((38,120,255,0), 0b01011101)

        expected = (37, 121, 255, 1)

        self.assertEqual(result, expected)



import unittest
import pytest
from src.image_handler import ImageHandler

class ImageHandlerTests(unittest.TestCase):
    def test_can_import_image(self):
        expected = {"filename": "./img/Tower_Bridge_from_Shad_Thames.jpg", "format": "JPEG", "mode": "RGB", "size": (1200, 600)}
        self.assertEqual(True, True)




from random import seed, randint

def get_pixels(width, height):
    for n in range(width):
        for m in range(height):
            yield (n, m)


def encode_data_to_pixel(pixel, data):
    new_pixel = [0, 0, 0, 0]

    data = bin(data)

    i = 0
    while i <= 3:
        channel = pixel[i]

        # Konverterar data till binärt i en sträng med korrekt antal siffror (8) 
        binary_data = f"0b{"0" * (10 - len(data))}{data[2:]}"
        
        # Plockar ut delen av datan att sätta i varje pixel
        subdata = int((binary_data)[2*i+2 : 2*i+4], 2)

        # Sätter in data
        channel = channel >> 2 << 2 
        channel += subdata
        
        new_pixel[i] = channel

        i += 1

    return tuple(new_pixel)


def decode_data_from_pixel(pixel):
    data = "0b"

    for channel in pixel:

        if len(bin(channel)) < 4:
            binary_data = f"0b0{bin(channel)[-1]}"
        else:
            binary_data = bin(channel)

        data += binary_data[-2:]

    data = int(data, 2)
    return data

  
def get_rail_fence_pixels(width, height, rail_fence_height):
    for n in range(width):
        for m in range(height):
            if m % (rail_fence_height * 2 - 2) == n % (rail_fence_height * 2 - 2):
                yield (n, m)

def get_random_spacing_pixels(width, height, key):
    seed(key)
    i = 0
    while i <= height:
        j = 0
        while j <= width:
            j += randint(1,5)
            yield(i, j)
        i += 1




import unittest
import pytest
from src import utils
class PixelDataDecodeTests(unittest.TestCase):
    def test_can_decode_pixel(self):

        result = utils.decode_data_from_pixel((37, 121, 255, 1))

        expected = 0b01011101

        self.assertEqual(result, expected)




def encode_rail_fence_cipher(message, rails):
    fence = {}
    for rail in range(rails):
        fence[rail] = ""
    rail = 0
    down = True
    for (idx, item) in enumerate(message):
        fence[rail] += item
        if down:
            rail += 1
        else:
            rail -= 1
        if rail % (rails - 1) == 0 :
            down = not down
    result = ""
    for value in fence.values():
        result += value
    return result
